import pandas as pd
import numpy as np
from typing import Callable
from rpy2 import robjects
import rpy2.robjects.packages as rpackages
from functools import cached_property

r_ssa = rpackages.importr('Rssa')
ssa_get = robjects.r('utils::getFromNamespace("$.ssa", "Rssa")')


class SSABase:

    def __init__(self, x, ssa_object, call: str):
        self.obj = ssa_object
        self._x = x
        self._call = call

    @cached_property
    def sigma(self):
        return np.asarray(ssa_get(self.obj, "sigma"))

    @cached_property
    def U(self):
        return np.asarray(ssa_get(self.obj, "U")).T

    @cached_property
    def V(self):
        return np.asarray(ssa_get(self.obj, "V"))

    @cached_property
    def series(self):
        return pd.DataFrame(self._x)

    def contributions(self, idx=None):
        if idx is None:
            idx = range(1, self.nsigma() + 1)
        return r_ssa.contributions(self.obj, idx)

    def nspecial(self):
        return r_ssa.nspecial(self.obj)[0]

    def nsigma(self):
        return r_ssa.nsigma(self.obj)[0]

    def nu(self):
        return r_ssa.nu(self.obj)[0]

    def __str__(self):
        final_call = self._call[:self._call.find('(') + 1] + "x=" + self._call[self._call.find('(') + 1:]

        result = str(self.obj).split("\n")
        result[result.index("Call:") + 1] = final_call
        return "\n".join(result)

    def __repr__(self):
        return self.__str__()


class SSA(SSABase):

    def __init__(self, x,
                 L=None,
                 neig=None,
                 mask=None,
                 wmask=None,
                 kind="1d-ssa",
                 circular=False,
                 column_projector="none",
                 row_projector="none",
                 svd_method="auto",
                 call=None):

        if L is None:
            L = (len(x) + 1) // 2

        if isinstance(x, pd.DataFrame):
            x = x.iloc[:, 0]

        self.L = L
        self.kind = kind
        super().__init__(x, r_ssa.ssa(x, L=L, neig=neig, mask=mask, wmask=wmask, kind=kind,
                                      circular=circular,
                                      column_projector=column_projector,
                                      row_projector=row_projector,
                                      svd_method=svd_method), call=call)


def _norm_conversion(func):
    def wrapper(x):
        return float(func(np.array(x)))

    return wrapper


def _default_norm(x):
    return np.sqrt(np.mean(x ** 2))


class IOSSA(SSABase):

    def __init__(self, x: SSA,
                 nested_groups,
                 tol=1e-5,
                 kappa=2,
                 maxiter=100,
                 norm: Callable = _default_norm,
                 trace=False,
                 kappa_balance=0.5,
                 call=None,
                 **kwargs):
        self.nested_groups = nested_groups
        self.tol = tol
        self.kappa = kappa
        self.maxiter = maxiter
        self.kappa_balance = kappa_balance
        self.trace = trace
        self.groups = nested_groups
        if norm is None:
            norm = _default_norm
        self.norm = robjects.rinterface.rternalize(_norm_conversion(norm))
        super().__init__(x.series, r_ssa.iossa(x=x, **{"nested.groups": self.nested_groups}, tol=self.tol,
                                               kappa=self.kappa, maxiter=self.maxiter, norm=self.norm,
                                               trace=self.trace,
                                               **{"kappa.balance": self.kappa_balance}, **kwargs), call=call)

    @property
    def iossa_groups(self):
        return self.groups

    def summary(self):
        result = str(self.obj).split("\n")
        result[result.index("Call:") + 1] = self._call
        return "\n".join(result)

    def __str__(self):
        return str(self.obj)

    def __repr__(self):
        return self.__str__()


class FOSSA(SSABase):

    def __init__(self, x: SSA,
                 nested_groups,
                 filter=(-1, 1),
                 gamma=np.inf,
                 normalize=True,
                 call=None,
                 **kwargs):
        self.nested_groups = nested_groups
        self.filter = filter
        self.gamma = gamma
        self.normalize = normalize
        super().__init__(x.series, r_ssa.fossa(x=x, **{"nested.groups": self.nested_groups}, filter=self.filter,
                                               gamma=self.gamma, normalize=self.normalize, **kwargs), call=call)

    def summary(self):
        result = str(self.obj).split("\n")
        result[result.index("Call:") + 1] = self._call
        return "\n".join(result)

    def __str__(self):
        return str(self.obj)

    def __repr__(self):
        return self.__str__()


import pandas as pd
import numpy as np
from typing import Literal
from rpy2 import robjects
import rpy2.robjects.packages as rpackages
from pyrssa.classes.SSA import SSABase
from pyrssa.classes.WCorMatrix import WCorMatrix
from rpy2.robjects import conversion

r_ssa = rpackages.importr('Rssa')
ssa_get = robjects.r('utils::getFromNamespace("$.ssa", "Rssa")')


class AutoGroup:

    def __init__(self, obj):
        self.obj = obj
        self.names = list(robjects.r.names(self.obj))
        self.groups = []
        for i, name in enumerate(self.names):
            group = np.asarray(self.obj.rx(name)[0])
            self.groups.append(group)
            setattr(self, name, group)

    def __getitem__(self, item):
        if isinstance(item, str):
            return getattr(self, item)
        elif isinstance(item, int):
            return self.groups[item]

    def keys(self):
        return self.names

    def values(self):
        return [self.__getitem__(name) for name in self.names]

    def items(self):
        return zip(self.keys(), self.values())

    def __len__(self):
        return len(self.names)

    def __str__(self):
        return str(self.obj)

    def __repr__(self):
        return self.__str__()


class GroupPgram(AutoGroup):

    def __init__(self, x: SSABase, groups=None, base: Literal["series", "eigen", "factor"] = "series", freq_bins=2,
                 threshold=0, method: Literal["constant", "linear"] = "constant", drop=True, **kwargs):
        if groups is None:
            groups = range(1, min(x.nsigma(), x.nu()) + 1)

        super().__init__(obj=r_ssa.grouping_auto_pgram(x=x, groups=groups, base=base, **{"freq.bins": freq_bins},
                                                       threshold=threshold, method=method, drop=drop, **kwargs))
        with conversion.localconverter(robjects.default_converter):
            contributions_obj = robjects.r.attr(self.obj, "contributions")
            self.contributions = pd.DataFrame(np.asmatrix(contributions_obj), columns=list(contributions_obj.colnames))

        self.type = robjects.r.attr(self.obj, "type")
        self.threshold = robjects.r.attr(self.obj, "threshold")


class GroupWCor(AutoGroup):

    def __init__(self, x: SSABase, groups=None, nclust=None, **kwargs):
        if groups is None:
            groups = range(1, min(x.nsigma(), x.nu()) + 1)
        if nclust is None:
            nclust = len(groups) // 2

        super().__init__(obj=r_ssa.grouping_auto_wcor(x=x, groups=groups, nclust=nclust, **kwargs))

        self.hclust = robjects.r.attr(self.obj, "hclust")
        self.wcor = WCorMatrix(np.asmatrix(robjects.r.attr(self.obj, "wcor")))



# import rpy2's package module
import rpy2.robjects.packages as rpackages
# R vector of strings
from rpy2.robjects.vectors import StrVector
# import R's utility package
utils = rpackages.importr('utils')
# select a mirror for R packages
utils.chooseCRANmirror(ind=1)  # select the first mirror in the list


def install_required():
    pack_names = ('Rssa', 'stats', )
    names_to_install = [x for x in pack_names if not rpackages.isinstalled(x)]
    if len(names_to_install) > 0:
        utils.install_packages(
            StrVector(names_to_install))



import pandas as pd
from rpy2 import robjects
from pyrssa.classes.SSA import SSA
import rpy2.robjects.packages as rpackages
from rpy2.robjects import conversion
from pyrssa.conversion import get_time_index, make_time_index
import numpy as np

r_ssa = rpackages.importr('Rssa')


class RForecast:

    def __init__(self, x: SSA, groups, length, base, only_new, reverse, drop, drop_attributes, cache, **kwargs):
        self.obj = r_ssa.rforecast(x, groups, **{"len": length}, base=base, **{"only.new": only_new}, reverse=reverse,
                                   drop=drop, **{"drop.attributes": drop_attributes}, cache=cache, **kwargs)

        self.names = robjects.r.names(self.obj)

        time_index = get_time_index(x.series)

        for name in self.names:
            series = pd.Series(self.obj.rx(name)[0])
            if time_index is not None:
                series.index = make_time_index(series, time_index, only_new=only_new)
            setattr(self, name, series)

    def __getitem__(self, item):
        if isinstance(item, str):
            return getattr(self, item)

    def __str__(self):
        return str(self.obj)

    def __repr__(self):
        return self.__str__()


class VForecast:

    def __init__(self, x: SSA, groups, length, only_new, drop, drop_attributes, **kwargs):
        self.obj = r_ssa.vforecast(x, groups, **{"len": length}, **{"only.new": only_new},
                                   drop=drop, **{"drop_attributes": drop_attributes}, **kwargs)

        self.names = robjects.r.names(self.obj)

        time_index = get_time_index(x.series)

        for name in self.names:
            series = pd.Series(self.obj.rx(name)[0])
            if time_index is not None:
                series.index = make_time_index(series, time_index, only_new=only_new)
            setattr(self, name, series)

    def __getitem__(self, item):
        if isinstance(item, str):
            return getattr(self, item)

    def __str__(self):
        return str(self.obj)

    def __repr__(self):
        return self.__str__()


class BForecast:

    def __init__(self, x: SSA, groups, length, r, level, kind, interval, only_new, only_intervals,
                 drop, drop_attributes, cache, **kwargs):
        self.obj = r_ssa.bforecast(x, groups, **{"len": length}, R=r, level=level, type=kind,
                                   interval=interval, **{"only.new": only_new}, **{"only.intervals": only_intervals},
                                   drop=drop, **{"drop.attributes": drop_attributes}, cache=cache, **kwargs)

        time_index = get_time_index(x.series)

        self.names = robjects.r.names(self.obj)
        with conversion.localconverter(robjects.default_converter):
            for name in self.names:
                forecast_df = pd.DataFrame(np.asmatrix(self.obj.rx2(name)), columns=list(self.obj.rx2(name).colnames))
                if time_index is not None:
                    forecast_df.index = make_time_index(forecast_df, time_index, only_new=only_new)
                setattr(self, name, forecast_df)

    def __getitem__(self, item):
        if isinstance(item, str):
            return getattr(self, item)

    def __str__(self):
        return str(self.obj)

    def __repr__(self):
        return self.__str__()



import os
# TODO: change behavior of searching for R (maybe by independently installing R in pyrssa).
if os.environ.get("R_HOME") is None:
    os.environ["R_HOME"] = r"C:\Program Files\R\R-4.2.1"


from pyrssa.classes.SSA import SSA, IOSSA, FOSSA
from pyrssa.classes.Parestimate import Parestimate
from pyrssa.classes.Resonstruction import Reconstruction
from pyrssa.classes.AutoSSA import GroupPgram, GroupWCor
from pyrssa.classes.Forecast import *
from pyrssa.classes.WCorMatrix import WCorMatrix
from pyrssa.classes.HMatrix import HMatrix
from pyrssa.base import *
from pyrssa.plot import plot, clplot



import numpy as np
from rpy2 import robjects
from pyrssa.classes.SSA import SSA
import rpy2.robjects.packages as rpackages
from functools import cached_property

r_ssa = rpackages.importr('Rssa')


class BaseParestimate:

    def __init__(self, par_obj=None, x=None, groups=None, method="esprit", subspace="column", normalize_roots=None,
                 dimensions=None, solve_method="ls", drop=True):
        if par_obj is not None:
            self.obj = par_obj
        else:
            self.obj = r_ssa.parestimate(x=x, groups=groups, method=method, subspace=subspace,
                                         normalize_roots=normalize_roots, dimensions=dimensions,
                                         solve_method=solve_method, drop=drop)

    @cached_property
    def roots(self):
        return np.asarray(self.obj.rx("roots")[0])

    @cached_property
    def periods(self):
        return np.asarray(self.obj.rx("periods")[0])

    @cached_property
    def frequencies(self):
        return np.asarray(self.obj.rx("frequencies")[0])

    @cached_property
    def moduli(self):
        return np.asarray(self.obj.rx("moduli")[0])

    @cached_property
    def rates(self):
        return np.asarray(self.obj.rx("rates")[0])

    def __getitem__(self, item):
        try:
            return getattr(self, item)
        except AttributeError:
            raise AttributeError(f"This parestimate group does not have attribute with name '{item}'.")

    def __str__(self):
        return str(self.obj)

    def __repr__(self):
        return self.__str__()


class Parestimate:

    def __init__(self, x, groups, method="esprit", subspace="column", normalize_roots=None,
                 dimensions=None, solve_method="ls", drop=True):
        if type(x) == SSA:
            method = "esprit"
            subspace = "column"
            normalize_roots = None
            dimensions = None
            solve_method = "ls"
            drop = True

        self.obj = r_ssa.parestimate(x=x, groups=groups, method=method, subspace=subspace,
                                     normalize_roots=normalize_roots, dimensions=dimensions,
                                     solve_method=solve_method, drop=drop)
        self.names = list(robjects.r.names(self.obj))

    def __getattribute__(self, item) -> BaseParestimate:
        try:
            return object.__getattribute__(self, item)
        except AttributeError:
            if item in self.names:
                res = BaseParestimate(par_obj=self.obj.rx(item)[0])
                setattr(self, item, res)
                return res
            else:
                raise AttributeError(f"This parestimate object does not have group with name '{item}'.")

    def __getitem__(self, item) -> BaseParestimate:
        if isinstance(item, str) and item in self.keys():
            return getattr(self, item)
        else:
            raise AttributeError(f"This parestimate object does not have group with name '{item}'.")

    def keys(self):
        return self.names

    def values(self):
        return [self.__getitem__(name) for name in self.names]

    def items(self):
        return zip(self.keys(), self.values())

    def __str__(self):
        return str(self.obj)

    def __repr__(self):
        return self.__str__()



from rpy2 import robjects
import rpy2.robjects.conversion as conversion
from rpy2.robjects import default_converter
from rpy2.robjects.numpy2ri import converter as numpy_converter
from rpy2.robjects.pandas2ri import converter as pandas_converter
import numpy as np
import pandas as pd
from pyrssa.classes.SSA import SSABase
from typing import Union


class FloatVector(robjects.FloatVector):
    def __init__(self, obj):
        super().__init__(obj)

    def __mul__(self, other):
        if isinstance(other, float):
            return robjects.r.sapply(self, "*", other)
        elif isinstance(other, int):
            return robjects.r.sapply(self, "*", other)

    def __truediv__(self, other):
        if isinstance(other, float):
            return robjects.r.sapply(self, "/", other)
        elif isinstance(other, int):
            return robjects.r.sapply(self, "/", other)

    def __add__(self, other):
        if isinstance(other, float):
            return robjects.r.sapply(self, "+", other)
        elif isinstance(other, int):
            return robjects.r.sapply(self, "+", other)

    def __pow__(self, power, modulo=None):
        if isinstance(power, float) or isinstance(power, int):
            return robjects.r.sapply(self, "^", power)

    __rmul__ = __mul__

    __radd__ = __add__


class IntVector(robjects.IntVector):
    def __init__(self, obj):
        super().__init__(obj)

    def __mul__(self, other):
        if isinstance(other, float):
            return FloatVector([i * other for i in self])

    def __truediv__(self, other):
        if isinstance(other, float):
            return robjects.r.sapply(self, "*", other)
        elif isinstance(other, int):
            return robjects.r.sapply(self, "*", other)

    def __pow__(self, power, modulo=None):
        if isinstance(power, float) or isinstance(power, int):
            return robjects.r.sapply(self, "^", power)

    __rmul__ = __mul__


def is_iterable(obj):
    try:
        iter(obj)
        return True
    except TypeError:
        return False


def is_list(obj):
    return is_iterable(obj) and not isinstance(obj, (str, dict))


def is_arr_of_type(arr, check_type, allow_inf=True, allow_na=True):
    if isinstance(arr, list) or isinstance(arr, tuple) or isinstance(arr, np.ndarray):
        for x in arr:
            if not isinstance(x, check_type):
                if not hasattr(x, '__iter__'):
                    if allow_na and not np.isnan(x) or allow_inf and not np.isinf(x):
                        return False
                else:
                    return False
        return True
    else:
        return False


def is_arr_of_types(obj, check_types, allow_inf=True, allow_na=True):
    if is_list(obj):
        return all([False if is_list(x) else
                    isinstance(x, check_types)
                    or (allow_na and np.isnan(x))
                    or (allow_inf and np.isinf(x)) for x in obj])


def is_int_arr(obj):
    return isinstance(obj, range) or is_arr_of_types(obj, (int, np.int32, np.int64))


def is_float_arr(obj):
    return is_arr_of_types(obj, (float, np.float32, np.float64))


def none_to_null(_):
    return robjects.r('NULL')


def range_to_vec(obj):
    return IntVector(list(obj))


def list_to_vec(obj):
    if is_int_arr(obj):
        return IntVector(list(obj))
    elif is_float_arr(obj):
        return FloatVector(list(obj))
    # elif is_of_int_lists_arr(obj):
    #     result = robjects.ListVector.from_length(len(obj))
    #     for i, x in enumerate(obj):
    #         if isinstance(x, int):
    #             result[i] = robjects.IntVector([x])
    #         else:
    #             result[i] = robjects.IntVector(list(x))
    #     return result
    else:
        return robjects.ListVector(list(obj))


def collection_conversion(obj):
    if is_list(obj):
        if is_int_arr(obj):
            return IntVector(list(obj))
        elif is_float_arr(obj):
            return FloatVector(list(obj))
        else:
            result = robjects.ListVector.from_length(len(obj))
            for i, x in enumerate(obj):
                result[i] = collection_conversion(x)
            return result
    elif isinstance(obj, dict):
        return robjects.vectors.ListVector({str(k): collection_conversion(obj[k]) for k in obj})
    else:
        return obj


def r_list(*args, **kwargs):
    if kwargs:
        for k in kwargs:
            if is_int_arr(kwargs[k]):
                kwargs[k] = list_to_vec(kwargs[k])
        return robjects.vectors.ListVector(kwargs)
    else:
        res = robjects.ListVector.from_length(len(args))
        for i in range(len(args)):
            res[i] = IntVector(args[i])
        return res


def dict_to_vec(obj):
    obj = {str(k): obj[k] for k in obj}
    return r_list(**obj)


def convert(obj):
    if is_int_arr(obj):
        return robjects.IntVector(obj)


def pyrssa_to_rssa(obj):
    return obj.obj


def get_time_index(series):
    if isinstance(series, pd.DataFrame) or isinstance(series, pd.Series):
        if isinstance(series.index, pd.DatetimeIndex):
            return series.index
    return None


def make_time_index(series: Union[pd.Series, pd.DataFrame], time_index: pd.DatetimeIndex, only_new=False):
    # if only_new is False, we have to ignore old series, when creating new indices
    periods = len(series) - (not only_new) * len(time_index) + 1
    new = pd.date_range(max(time_index), freq=time_index.freqstr, periods=periods,
                        inclusive="right")
    if only_new:
        return new
    else:
        return time_index.union(new)


pyrssa_converter = conversion.Converter('pyrssa converter')
pyrssa_converter.py2rpy.register(type(None), none_to_null)
pyrssa_converter.py2rpy.register(range, collection_conversion)
pyrssa_converter.py2rpy.register(tuple, collection_conversion)
pyrssa_converter.py2rpy.register(list, collection_conversion)
pyrssa_converter.py2rpy.register(dict, collection_conversion)
pyrssa_converter.py2rpy.register(SSABase, pyrssa_to_rssa)

# pyrssa_conversion_rules = default_converter + pyrssa_converter + numpy_converter + pandas_converter
pyrssa_conversion_rules = pandas_converter + numpy_converter + default_converter + pyrssa_converter



import rpy2.robjects.packages as rpackages
import numpy as np
from functools import cached_property

r_stats = rpackages.importr('stats')


class Periodogram:

    def __init__(self, x, spans=None, kernel=None, taper=0.1,
                 pad=0, fast=True, demean=False, detrend=True, **kwargs):

        self.obj = r_stats.spec_pgram(x, spans=spans, kernel=kernel, taper=taper, pad=pad,
                                      fast=fast, demean=demean, detrend=detrend, plot=False, **kwargs)
        self.series = x
        self.detrend = detrend
        self.demean = demean
        self.taper = taper
        self.pad = pad

    @cached_property
    def freq(self):
        return np.asarray(self.obj.rx("freq")[0])

    @cached_property
    def spec(self):
        return np.asarray(self.obj.rx("spec")[0])

    @cached_property
    def coh(self):
        return np.asarray(self.obj.rx("coh")[0])

    @cached_property
    def phase(self):
        return np.asarray(self.obj.rx("phase")[0])

    @cached_property
    def kernel(self):
        return np.asarray(self.obj.rx("kernel")[0])

    @cached_property
    def df(self):
        return self.obj.rx("df")[0][0]

    @cached_property
    def bandwidth(self):
        return self.obj.rx("bandwidth")[0][0]

    @cached_property
    def n_used(self):
        return self.obj.rx("n.used")[0][0]

    @cached_property
    def orig_n(self):
        return self.obj.rx("orig.n")[0][0]

    @cached_property
    def snames(self):
        return np.asarray(self.obj.rx("snames")[0])

    @cached_property
    def method(self):
        return self.obj.rx("method")[0][0]

    def __str__(self):
        return str(self.obj)

    def __repr__(self):
        return self.__str__()



import pandas as pd
import numpy as np
from rpy2 import robjects
from pyrssa.classes.SSA import SSABase
from pyrssa.classes.AutoSSA import GroupPgram, GroupWCor
from pyrssa.conversion import get_time_index
import rpy2.robjects.packages as rpackages
from typing import Union
from functools import cached_property

r_ssa = rpackages.importr('Rssa')


class Reconstruction:

    def __init__(self, x: SSABase, groups: Union[list, dict, np.ndarray, GroupPgram, GroupWCor],
                 drop_attributes=False, cache=True):
        if isinstance(groups, GroupPgram) or isinstance(groups, GroupWCor):
            groups = dict(zip(groups.names, groups.groups))
        self.obj = r_ssa.reconstruct(x=x, groups=groups, **{"drop.attributes": drop_attributes}, cache=cache)
        self._x = x
        self.names = list(robjects.r.names(self.obj))
        self._datetime_index = get_time_index(x.series)

    @cached_property
    def series(self):
        return self._x.series

    @cached_property
    def residuals(self):
        return np.asarray(robjects.r.attr(self.obj, "residuals"))

    def __getattribute__(self, item):
        try:
            return object.__getattribute__(self, item)
        except AttributeError:
            if item in self.names:
                series = pd.Series(self.obj.rx(item)[0])
                series.name = item
                if self._datetime_index is not None:
                    series.index = self._datetime_index
                setattr(self, item, series)
                return series
            else:
                raise AttributeError

    def __getitem__(self, item):
        if isinstance(item, str) and item in self.keys():
            return getattr(self, item)
        else:
            raise AttributeError(f"This reconstruction object does not have series with name '{item}'.")

    def keys(self):
        return self.names

    def values(self):
        return [self.__getitem__(name) for name in self.names]

    def items(self):
        return zip(self.keys(), self.values())

    def __str__(self):
        return str(self.obj)

    def __repr__(self):
        return self.__str__()



import pandas as pd

from rpy2 import robjects
from pyrssa.classes.SSA import SSABase
from pyrssa.classes.Periodogram import Periodogram
from pyrssa.conversion import is_list
from pyrssa import Reconstruction, reconstruct
from pyrssa import WCorMatrix, HMatrix
from pyrssa import GroupPgram
from pyrssa import calc_v
import matplotlib
import matplotlib.pyplot as plt
import matplotlib.gridspec as gridspec
import numpy as np
from typing import Literal, Union

# This line solves problem for PyCharm: module 'backend_interagg' has no attribute 'FigureCanvas'...
matplotlib.use('TkAgg')
_num_complete = robjects.r('utils::getFromNamespace("num.complete", "Rssa")')


def _should_share_limits(series_arr, max_diff=1, max_mul=1.5):
    range_values = [[min(ser), max(ser), max(ser) - min(ser)] for ser in series_arr]
    arr_start, arr_end, arr_range = list(map(list, zip(*range_values)))
    min_start = min(arr_start)
    max_end = max(arr_end)
    max_start = max(arr_start)
    min_end = min(arr_end)
    max_range = abs(min_start - max_end)
    return abs(max_start - min_start) / max_range <= max_diff \
           and abs(max_end - min_end) / max_range <= max_diff \
           and max_range / max(arr_range) <= max_mul


class Plot:

    @staticmethod
    def get_kwargs(kwargs: dict):
        return {key: value.pop(0) if is_list(value) else value if value else None for key, value in kwargs.items()}

    @staticmethod
    def set_style(style_name="seaborn-v0_8-whitegrid"):
        plt.style.use(style_name)

    @staticmethod
    def bold(text: str):
        return r"$\bf{" + text.replace(' ', r'\ ') + "}$"

    @staticmethod
    def vectors(x: SSABase, vectors: Literal["eigen", "factor"] = "eigen",
                idx=None, contrib=True, layout=None, title=None, show=True):

        if idx is None:
            idx = range(1, min(10, len(x.U)) + 1)
        if contrib is True:
            cntrb = x.contributions(idx)
        else:
            cntrb = None

        if vectors == "eigen":
            vec, main_title = x.U, "Eigenvectors"
        else:
            vec, main_title = calc_v(x, idx=idx), "Factor vectors"

        if layout is None:
            cols = 4
            rows = int(np.ceil(len(idx) / cols))
        else:
            rows = layout[0]
            cols = layout[1]

        fig = plt.figure(figsize=(cols + 2, rows + 2))
        gs = gridspec.GridSpec(rows, cols, width_ratios=[1] * cols, figure=fig, wspace=0, hspace=0.5)
        fig.suptitle(main_title if title is None else title, fontweight="bold")
        fig.tight_layout()
        ax = None

        for i in range(len(idx)):
            ax = fig.add_subplot(gs[i], sharey=ax)
            ax.plot(vec[idx[i] - 1], linewidth=0.5)
            ax.set_title(f'{idx[i]} ({cntrb[i] * 100:.2f} %)' if cntrb is not None else idx[i])
            ax.set_xticks([])
            ax.set_yticks([])
            ax.set_box_aspect(1)

        if show:
            plt.show()
        return plt

    @staticmethod
    def paired(x: SSABase, vectors: Literal["eigen", "factor"] = "eigen",
               idx=None, contrib=True, layout=None, title=None, show=True):

        if idx is None:
            idx = range(len(x.U))
        if contrib is True:
            cntrb = x.contributions(idx)
        else:
            cntrb = None

        if vectors == "eigen":
            vec, main_title = x.U, "Pairs of eigenvectors"
        else:
            vec, main_title = calc_v(x, idx=idx), "Pairs of factor vectors"

        if layout is None:
            cols = 4
            rows = int(np.ceil((len(idx) - 1) / cols))
        else:
            rows = layout[0]
            cols = layout[1]

        fig = plt.figure(figsize=(cols + 2, rows + 2))
        gs = gridspec.GridSpec(rows, cols, width_ratios=[1] * cols, figure=fig, wspace=0, hspace=0.5)
        fig.suptitle(main_title if title is None else title, fontweight="bold")
        fig.tight_layout()
        # fig.subplots_adjust(top=0.8) --> maybe remove after 2 month (from 07.04.23)

        for i in range(len(idx) - 1):
            ax = fig.add_subplot(gs[i])
            ax.plot(vec[idx[i + 1] - 1], vec[idx[i] - 1], linewidth=0.25)
            if cntrb is None:
                ax.set_title(f'{idx[i]} vs {idx[i + 1]}')
            else:
                ax.set_title(f'{idx[i]} ({cntrb[i] * 100:.2f}%) vs {idx[i + 1]} ({cntrb[i + 1] * 100:.2f}%)')
            # lim = np.min(vec[idx[i] - 1:idx[i + 1]]), np.max(vec[idx[i] - 1:idx[i + 1]])

            xlim = np.min(vec[idx[i + 1] - 1]), np.max(vec[idx[i + 1] - 1])
            ylim = np.min(vec[idx[i] - 1]), np.max(vec[idx[i] - 1])
            if i == 0:
                print(xlim, ylim)
            x_shift = abs(xlim[1] - xlim[0]) * 0.05
            y_shift = abs(ylim[1] - ylim[0]) * 0.05
            xlim = xlim[0] - x_shift, xlim[1] + x_shift
            ylim = ylim[0] - y_shift, ylim[1] + y_shift
            ax.set_xticks([])
            ax.set_yticks([])
            ax.set_xlim(xlim)
            ax.set_ylim(ylim)
            ax.set_box_aspect(1)

        if show:
            plt.show()
        return plt

    def series(self, x: SSABase, groups=None, layout=None, **kwargs):
        return self.xyplot(reconstruct(x, groups=groups),
                           add_residuals=False, add_original=False,
                           layout=layout, **kwargs)

    def xyplot(self, x: Reconstruction, x_labels=None, add_residuals=True, add_original=True,
               layout=None, superpose=False, title=None, show=True, **kwargs):

        if not isinstance(x, Reconstruction):
            raise TypeError(f"Only Reconstruction type of object is allowed in xyplot. You've tried to pass {type(x)}.")

        if x_labels is None:
            x_labels = x.series.index

        kwargs.setdefault("linewidth", 0.75)

        if superpose:
            fig, ax = plt.subplots()

            if add_original:
                ax.plot(x_labels, x.series, label="Original", **self.get_kwargs(kwargs))
            for name in x.names:
                ax.plot(x_labels, getattr(x, name), label=name, **self.get_kwargs(kwargs))
            if add_residuals:
                ax.plot(x_labels, x.residuals, label='Residuals', **self.get_kwargs(kwargs))
            ax.legend()
            plt.title(label=self.bold("Reconstructed Series" if title is None else title))
            fig.tight_layout()

        else:

            cnt = len(x.names) + add_original + add_residuals

            if layout is not None:
                if layout[0] * layout[1] < cnt:
                    raise ValueError(f"Layout size {layout} is less than count of plotting series ({cnt}).")
                rows = layout[0]
                cols = layout[1]
            else:
                rows = cnt
                cols = 1

            plotting_series = []
            if add_original:
                plotting_series.append({'name': 'Original', 'series': x.series})
            for i in range(add_original, cnt - add_residuals):
                name = x.names[i - add_original]
                plotting_series.append({'name': name, 'series': getattr(x, name)})
            if add_residuals:
                plotting_series.append({'name': 'Residuals', 'series': x.residuals})

            if cols > 1:
                all_series = [ser['series'] for ser in plotting_series]
                share_y = _should_share_limits(all_series)
            else:
                share_y = False

            fig = plt.figure(figsize=(2 * (cols + 2), 2 * (rows + 2)))
            if share_y:
                gs = gridspec.GridSpec(rows, cols, width_ratios=[1] * cols, figure=fig, wspace=0.1, hspace=0.5)
            else:
                gs = gridspec.GridSpec(rows, cols, width_ratios=[1] * cols, figure=fig, hspace=0.7)
            fig.suptitle("Reconstructed Series", fontweight="bold")
            ax = None

            for i in range(len(plotting_series)):
                if share_y:
                    is_first = ax is None
                    ax = fig.add_subplot(gs[i], sharex=ax, sharey=ax)
                    if not is_first:
                        ax.tick_params(left=False, labelleft=False)
                else:
                    ax = fig.add_subplot(gs[i], sharex=ax)

                ax.plot(x_labels, plotting_series[i]['series'], **self.get_kwargs(kwargs))
                if len(plotting_series) > 1:
                    ax.set_title(plotting_series[i]['name'])

        if show:
            plt.show()
        return plt

    def matplot(self, x: Reconstruction, x_labels=None, add_residuals=True,
                add_original=True, title=None, show=True, **kwargs):

        if x_labels is None:
            x_labels = x.series.index

        fig, ax = plt.subplots()
        if add_original:
            ax.plot(x_labels, x.series, label="Original", color="black", linewidth=0.5, **self.get_kwargs(kwargs))
        for name in x.names:
            ax.plot(x_labels, getattr(x, name), label=name, linestyle="dashed", **self.get_kwargs(kwargs))
        if add_residuals:
            ax.plot(x_labels, x.residuals, label='Residuals', linestyle="dashed", **self.get_kwargs(kwargs))
        ax.legend()
        plt.title(label=self.bold("Reconstructed Series" if title is None else title))
        if show:
            plt.show()
        return plt

    @staticmethod
    def sigma(ts: SSABase, show=True):
        plt.suptitle('Component norms', fontweight="bold")
        plt.plot(ts.sigma, marker='o')
        plt.xlabel('Index')
        plt.ylabel('Norms')
        plt.yscale('log')
        if show:
            plt.show()
        return plt

    def _wcor(self, wcor_matrix: WCorMatrix, scales=None, show=True):
        plt.imshow(wcor_matrix, cmap='gray_r', vmin=0, vmax=1)
        plt.gca().invert_yaxis()
        plt.grid(color='k', alpha=0.2, linestyle='-', linewidth=0.3)
        if scales is None:
            ticks = range(len(wcor_matrix.groups))
            labels = wcor_matrix.groups
        else:
            ticks = np.array(scales) - 1  # fix for indexing of components on the plot
            labels = scales
        plt.title(self.bold("W-correlation matrix"))
        plt.xticks(ticks, labels=labels)
        plt.yticks(ticks, labels=labels)

        minor_cnt = wcor_matrix.shape[0]
        plt.xticks(np.arange(-.5, minor_cnt, 1), minor=True)
        plt.yticks(np.arange(-.5, minor_cnt, 1), minor=True)

        plt.grid(which='minor', color='w', linestyle='-', linewidth=0.2)
        plt.tick_params(which='minor', bottom=False, left=False)

        if show:
            plt.show()
        return plt

    def _hmatrix(self, hmatrix: HMatrix, show=True):
        plt.imshow(hmatrix.T, cmap='hot_r', origin='lower', interpolation='nearest')
        plt.title(self.bold("Heterogeneity matrix"))
        if show:
            plt.show()
        return plt

    def _group_pgram(self, x: GroupPgram, order=False, legend_params: dict = None, show=True, **kwargs):
        contrib = x.contributions
        if order:
            contrib = contrib.transform(np.sort)[::-1]
            contrib.reset_index(drop=True, inplace=True)
        for column in contrib:
            plt.plot(contrib[column], label=column, **self.get_kwargs(kwargs))
        if legend_params is None:
            legend_params = {}
        legend_params.setdefault("loc", "lower center")
        legend_params.setdefault("bbox_to_anchor", (0.5, 1.05))
        plt.xlabel("Component")
        plt.ylabel("Relative contribution")
        plt.legend(**legend_params)
        plt.tight_layout()
        if show:
            plt.show()
        return plt

    def pgram(self, objs: Union[Periodogram, list], ticks=None, tick_labels=None, limits=None, show=True, **kwargs):
        if is_list(objs) and all([isinstance(obj, Periodogram) for obj in objs]):
            all_series = objs
        elif isinstance(objs, Periodogram):
            all_series = [objs]
        else:
            raise TypeError(f"Only Periodogram or list of periodograms are available as plotting objects,"
                            f" not {type(objs)}")

        kwargs.setdefault("linewidth", 0.75)
        fig, ax = plt.subplots()
        for i, pgram in enumerate(all_series):
            name = pgram.series.name if pgram.series.name is not None else f"Series {i + 1}"
            ax.plot(pgram.freq, pgram.spec, label=name, **self.get_kwargs(kwargs))
        if len(all_series) > 1:
            plt.title(self.bold("Spectrum of series"))
            plt.legend()
        else:
            plt.title(self.bold(f"Spectrum of {all_series[0].series.name}"))

        if ticks:
            ticks = [tick * 2 for tick in ticks]
            if tick_labels:
                plt.xticks(ticks, labels=tick_labels)
            else:
                plt.xticks(ticks)

        if limits is None:
            limits = (0, 0.5)
        shift = abs(limits[1] - limits[0]) * 0.05
        limits = (limits[0] - shift, limits[1] + shift)
        plt.xlim(limits)
        plt.ylabel("Value")
        if show:
            plt.show()
        return plt

    def __call__(self, obj, x_labels=None, kind: Literal["vectors", "paired"] = None,
                 add_residuals=True, add_original=True, idx=None, scales=None,
                 vectors: Literal["eigen", "factor"] = "eigen", contrib=True, layout=None, superpose=False,
                 method: Literal["matplot", "xyplot"] = "xyplot", title=None, groups=None, order=False,
                 legend_params: dict = None, ticks=None, tick_labels=None, limits=None, show=True, **kwargs):
        if kind == "vectors":
            return self.vectors(obj, vectors=vectors, idx=idx, contrib=contrib, layout=layout, title=title, show=show)
        elif kind == "paired":
            return self.paired(obj, vectors=vectors, idx=idx, contrib=contrib, layout=layout, title=title, show=show)
        elif kind == "series":
            return self.series(obj, groups=groups, layout=layout, show=show, **kwargs)
        elif isinstance(obj, Reconstruction):
            if method == "matplot":
                return self.matplot(obj, x_labels=x_labels, add_residuals=add_residuals, add_original=add_original,
                                    title=title, show=show, **kwargs)
            if method == "xyplot":
                return self.xyplot(obj, x_labels=x_labels, add_residuals=add_residuals, add_original=add_original,
                                   layout=layout, superpose=superpose, title=title, show=show, **kwargs)
        elif isinstance(obj, SSABase):
            return self.sigma(obj, show=show)
        elif isinstance(obj, WCorMatrix):
            return self._wcor(obj, scales=scales, show=show)
        elif isinstance(obj, HMatrix):
            return self._hmatrix(obj, show=show)
        elif isinstance(obj, GroupPgram):
            return self._group_pgram(obj, order=order, legend_params=legend_params, show=show, **kwargs)
        elif isinstance(obj, Periodogram) or isinstance(obj, list) and all([isinstance(x, Periodogram) for x in obj]):
            return self.pgram(obj, ticks=ticks, tick_labels=tick_labels, limits=limits, show=show, **kwargs)


def clplot(x, show=True):
    if isinstance(x, pd.DataFrame):
        x = x[x.columns[0]]
    if isinstance(x, pd.Series):
        x = x.values
    N = len(x)
    na_idx = np.hstack(np.argwhere(np.isnan(x))) + 1
    cr = [_num_complete(N, L=L, **{"na.idx": na_idx})[0] / (N - L + 1) * 100 for L in range(2, (N + 1) // 2 + 1)]
    plt.plot(cr)
    plt.title(plot.bold("Proportion of complete lag vectors"))
    plt.ylabel("Percents")
    plt.xlabel("Window length, L")
    if show:
        plt.show()
    return plt


plot = Plot()
plot.set_style()



import rpy2.robjects.packages as rpackages
import numpy as np

r_ssa = rpackages.importr('Rssa')


class HMatrix(np.matrix):

    def __new__(cls, F, B=None, T=None, L=None, neig=10):
        n = len(F)
        if B is None:
            B = n // 4
        if T is None:
            T = n // 4
        if L is None:
            L = B // 2
        obj = np.asmatrix(r_ssa.hmatr(F=F, B=B, T=T, L=L, neig=neig).view(cls))
        return obj

    def __array_finalize__(self, obj):
        pass



from setuptools import setup

setup(
    name='pyrssa',
    packages=['pyrssa', 'pyrssa.classes'],
    include_package_data=True,
    package_data={
        'data': ['*']
    },
    version='1.0.9',
    description='Rssa for Python',
    author='Fleyderer',
    author_email='fleyderer@gmail.com',
    url="https://github.com/Fleyderer/pyrssa",
    license="Apache 2.0",
    install_requires=['rpy2', 'pandas', 'numpy', 'matplotlib']
)



from rpy2 import robjects
import rpy2.robjects.packages as rpackages
import numpy as np

r_ssa = rpackages.importr('Rssa')


class WCorMatrix(np.matrix):

    def __new__(cls, x, groups=None, **kwargs):

        obj = None

        if groups is None:

            if isinstance(x, np.matrix):
                groups = range(1, min(*x.shape) + 1)
                obj = x
            else:
                groups = range(1, min(x.nsigma(), x.nu()) + 1)

        if obj is None:
            obj = np.asmatrix(robjects.r.wcor(x=x, groups=groups, **kwargs)).view(cls)

        obj.groups = np.copy(groups)
        return obj

    def __array_finalize__(self, obj):
        if obj is None:
            return
        self.groups = getattr(obj, 'groups', None)



from pyrssa.classes.SSA import SSABase
from pyrssa.classes.Parestimate import BaseParestimate
from pyrssa.classes.Periodogram import Periodogram
from pyrssa import SSA, IOSSA, FOSSA, Parestimate
from pyrssa import Reconstruction
from pyrssa import RForecast, VForecast, BForecast
from pyrssa import WCorMatrix, HMatrix
from pyrssa import GroupPgram, GroupWCor
from pyrssa import installer
from pyrssa.conversion import pyrssa_conversion_rules
from rpy2 import robjects
import rpy2.robjects.conversion as conversion
import rpy2.robjects.packages as rpackages
from rpy2.rinterface_lib import callbacks
import pandas as pd
import numpy as np
import os
from typing import overload, Literal, Union
import inspect

# Set conversion rules
conversion.set_conversion(pyrssa_conversion_rules)

# Ignore warnings
callbacks.consolewrite_warnerror = lambda *args: None

# Install required R packages.
installer.install_required()

r = robjects.r
r_ssa = rpackages.importr('Rssa')


def _get_call(frame):
    call = inspect.getframeinfo(frame)[3][0]
    return call[call.find('=') + 1:].strip().rstrip()


def _set_datetime_index(dataframe, name):
    dataframe[name] = pd.DatetimeIndex(dataframe[name])
    dataframe.set_index(name, inplace=True, drop=True)
    dataframe.index.freq = dataframe.index.inferred_freq


# Read pyrssa dataframes
def data(name, datetime_index=None):
    """
    Function for loading available in pyrssa package datasets. Available datasets are stored in the data directory.

    :param name: Name of dataset to load
    :param datetime_index: Name of column, where date or time index is stored
    :return: Loaded dataset
    :rtype: pandas.DataFrame or pandas.Series

    """
    result = pd.read_csv(os.path.join(os.path.join(os.path.dirname(os.path.realpath(__file__)), "data"), f'{name}.csv'))
    if datetime_index is not None:
        _set_datetime_index(result, datetime_index)
    elif len(result.columns) > 1:
        # Search for datetime index and set it
        col_idx = next((i for i, v in enumerate(result.columns) if v.lower() in ['time', 'date', 'datetime']), None)
        if col_idx is not None:
            _set_datetime_index(result, result.columns[col_idx])

    # If there is only one column in dataframe, return it as series
    if len(result.columns) == 1:
        return result[result.columns[0]]
    else:
        return result


def calc_v(x: SSABase, idx, **kwargs):
    return np.asarray(r_ssa.calc_v(x=x, idx=idx, **kwargs)).T


def parestimate(x, groups, method="esprit", subspace="column", normalize_roots=None, dimensions=None,
                solve_method="ls", drop=True):
    """
    Function to estimate the parameters (frequencies and rates) given a set of SSA eigenvectors.

    :param x: SSA object
    :param groups: list of indices of eigenvectors to estimate from
    :param method: For 1D-SSA,
        Toeplitz SSA, and MSSA: parameter estimation method, 'esprit' for 1D-ESPRIT (Algorithm 3.3 in Golyandina et al.
        (2018)), 'pairs' for rough estimation based on pair of eigenvectors (Algorithm 3.4 in Golyandina et al (2018)).
        For nD-SSA: parameter estimation method. For now only 'esprit' is supported (Algorithm 5.6 in Golyandina et al.
        (2018)). lowest dimension, when possible (length of groups is one)
    :param subspace: which subspace will be used for parameter estimation
    :param normalize_roots: logical vector or None, force signal roots to lie on unit circle.
        None means automatic selection: normalize iff circular topology OR Toeplitz SSA used
    :param dimensions: a vector of dimension indices to perform ESPRIT along. None means all dimensions.
    :param solve_method: approximate matrix equation solving method, 'ls' for least-squares,
        'tls' for total-least-squares.
    :param drop: logical, if 'TRUE' then the result is coerced to the lowest dimension,
        when possible (length of groups is one)

    :return:

    """
    if len(groups) == 1:
        return BaseParestimate(x=x, groups=groups, method=method, subspace=subspace,
                               normalize_roots=normalize_roots, dimensions=dimensions,
                               solve_method=solve_method, drop=drop)
    else:
        return Parestimate(x=x, groups=groups, method=method, subspace=subspace,
                           normalize_roots=normalize_roots, dimensions=dimensions,
                           solve_method=solve_method, drop=drop)


def ssa(x, L=None, neig=None, mask=None, wmask=None, kind="1d-ssa", circular=False,
        column_projector="none", row_projector="none", svd_method="auto"):
    """

    :param x: object to be decomposed. If DataFrame passed, the first column will be treated as a series
    :type x: pandas.DataFrame, pandas.Series, numpy.ndarray, list
    :param L: window length. Fixed to half of the series length by default.
        Should be vector of length 2 for 2d SSA
    :type L: int, optional
    :param neig: number of desired eigentriples. If None, then sane default value
        will be used.
    :type neig: int, optional
    :param mask: for shaped 2d SSA case only. Logical matrix with same dimension as x.
        Specifies form of decomposed array. If None, then all non-NA elements will be used
    :param wmask: for shaped 2d SSA case only. Logical matrix which specifies window form.
    :param kind: SSA method. This includes ordinary 1d SSA, 2d SSA, Toeplitz variant of 1d SSA, multichannel
        variant of SSA and complex SSA. Defaults to 1d SSA.
    :type kind: str, optional
    :param circular: logical vector of one or two elements, describes series topology for 1d SSA and Toeplitz SSA
        or field topology for 2d SSA. 'TRUE' means series circularity for 1d case or circularity
        by a corresponding coordinate for 2d case. See Shlemov and Golyandina (2014) for more information.
    :param column_projector, row_projector: column and row signal subspaces projectors for SSA with projection.
    :type column_projector: str or int, optional
    :type row_projector: str or int, optional
    :param svd_method: 	singular value decomposition method.
    :return: SSA object. The precise layout of the object is mostly meant opaque and subject to
        change in different version of the package.
    :rtype: SSA

    Description
    ===========

    Set up the SSA object and perform the decomposition, if necessary.

    Details
    ===========

    This is the main entry point to the package. This routine constructs the SSA object filling all necessary
    internal structures and performing the decomposition if necessary. For the comprehensive description of SSA
    modifications and their algorithms see Golyandina et al. (2018).

    Variants of SSA
    ---------------

    The following implementations of the SSA method are supported (corresponds to different values of kind argument):

    * 1d-ssa

        Basic 1d SSA as described in Chapter 1 of Golyandina et al. (2001). This is also known as Broomhead-King
        variant of SSA or BK-SSA, see Broomhead and King (1986).

    * toeplitz-ssa

        Toeplitz variant of 1d SSA. See Section 1.7.2 in Golyandina et al. (2001). This is also known
        as Vautard-Ghil variant of SSA or VG-SSA for analysis of stationary time series, see Vautard and Ghil (1989).

    * mssa

        Multichannel SSA for simultaneous decomposition of several time series (possible of unequal length). See
        Golyandina and Stepanov (2005).

    * cssa

        Complex variant of 1d SSA.

    * 2d-ssa

        2d SSA for decomposition of images and arrays. See Golyandina and Usevich (2009) and Golyandina et al. (2015)
        for more information.

    * nd-ssa

        Multidimensional SSA decomposition for arrays (tensors).

    Window shape selection (for shaped 2d SSA)
    ------------------------------------------

    Window shape may be specified by argument wmask. If wmask is 'NULL', then standard rectangular window (specified
    by L) will be used.

    * circle(R)

        circular mask of radius R

    * triangle(side)

        mask in form of isosceles right-angled triangle with cathetus side. Right angle lay on top left corner of
        container square matrix

    Also in wmask one may use following functions:

    These functions are not exported, they defined only for wmask expression. If one has objects with the same names
    and wants to use them rather than these functions, one should use special wrapper function I() (see 'Examples').

    Projectors specification for SSA with projection
    ------------------------------------------------

    Projectors are specified by means of column.projector and row.projector arguments (see Golyandina and Shlemov (
    2017)). Each may be a matrix of orthonormal (otherwise QR orthonormalization process will be performed) basis of
    projection subspace, or single integer, which will be interpreted as dimension of orthogonal polynomial basis (
    note that the dimension equals to degree plus 1, e.g. quadratic basis has dimension 3), or one of following
    character strings (or unique prefix): 'none', 'constant' (or 'centering'), 'linear', 'quadratic' or 'qubic' for
    orthonormal bases of the corresponding functions.

    Here is the list of the most used options

    * both projectors are 'none'

        corresponds to ordinary 1D SSA,

    * column.projector='centering'

        corresponds to 1D SSA with centering,

    * column.projector='centering' and row.projector='centering'

        corresponds to 1D SSA with double centering.

    SSA with centering and double centering may improve the separation of linear trend (see Golyandina et al. (2001)
    for more information).

    SVD methods
    -----------

    The main step of the SSA method is the singular decomposition of the so-called series trajectory matrix. Package
    provides several implementations of this procedure (corresponds to different values of svd.method) argument:

    * auto

        Automatic method selection depending on the series length, window length, SSA kind and number of eigenvalues
        requested.

    * nutrlan

        Thick-restart Lanczos eigensolver which operates on cross-product matrix. These methods exploit the Hankel
        structure of the trajectory matrix efficiently and is really fast. The method allows the truncated SVD (only
        specified amount of eigentriples to be computed) and the continuation of the decomposition. See Korobeynikov (
        2010) for more information.

    * propack

        SVD via implicitly restarted Lanczos bidiagonalization with partial reorthogonalization. These methods exploit
        the Hankel structure of the trajectory matrix efficiently and is really fast. This is the 'proper' SVD
        implementation (the matrix of factor vectors are calculated), thus the memory requirements of the methods are
        higher than for nu-TRLAN. Usually the method is slightly faster that nu-TRLAN and more numerically stable.
        The method allows the truncated SVD (only specified amount of eigentriples to be computed). See Korobeynikov (
        2010) for more information.

    * svd

        Full SVD as provided by LAPACK DGESDD routine. Neither continuation of the decomposition nor the truncated
        SVD is supported. The method does not assume anything special about the trajectory matrix and thus is slow.

    * eigen

        Full SVD via eigendecompsition of the cross-product matrix. In many cases faster than previous method,
        but still really slow for more or less non-trivial matrix sizes.

    * rspectra

        SVD via svds function from Rspectra package (if installed)

    * primme

        SVD via svds function from PRIMME package (if installed)

    Usually the ssa function tries to provide the best SVD implementation for given series length and the window
    size. In particular, for small series and window sizes it is better to use generic black-box routines (as
    provided by 'svd' and 'eigen' methods). For long series special-purpose routines are to be used.

    Examples
    --------

    .. code-block:: python

        import pyrssa as prs
        import pandas as pd
        import numpy as np

        AustralianWine = prs.data("AustralianWine")
        fort = AustralianWine['Fortified'][:174]
        fort.index = pd.date_range(start='1980/01/01', freq='M', periods=len(fort))
        s_fort = prs.ssa(fort, L=84, kind="1d-ssa")

        co2 = data("co2")
        s2 = prs.ssa(co2.value, column_projector="centering", row_projector="centering")
        s4 = prs.ssa(co2.value, column_projector=2, row_projector=2)

        S = np.exp(a * np.arange(1, n + 1)) * np.sin(2 * np.pi * np.arange(1, n + 1) / 7)
        f = S + sigma * np.random.normal(size=n)
        f_center = f - np.mean(f)
        s = prs.ssa(f, L=L, kind="1d-ssa")
        st = prs.ssa(f_center, L=L, kind='toeplitz-ssa')

    """
    return SSA(x, L=L, neig=neig, mask=mask, wmask=wmask, kind=kind,
               circular=circular,
               column_projector=column_projector,
               row_projector=row_projector,
               svd_method=svd_method,
               call=_get_call(inspect.currentframe().f_back))


def reconstruct(x: SSABase,
                groups: Union[list, dict, np.ndarray, GroupPgram, GroupWCor],
                drop_attributes=False,
                cache=True):
    """

    :param x: SSA object
    :type x: SSA
    :param groups: list of numeric vectors, indices of elementary components used for reconstruction,
        the entries of the list can be named, see 'Value' for more information
    :type groups: list or dict
    :param drop_attributes: if `True` then the attributes of the input objects are not copied to the reconstructed ones.
    :type drop_attributes: bool
    :param cache: if `True` then intermediate results will be cached in the SSA object.
    :type cache: bool
    :return: List of reconstructed objects. Elements of the list have the same names as elements of groups. If the
        group is unnamed, then corresponding component will obtain name ‘Fn’, where ‘n’ is its index in groups list.
    :rtype: Reconstruction

    Description
    ===========

    Reconstruct the data given the SSA decomposition and the desired grouping of the elementary components.

    Details
    =======

    Reconstruction is performed in a common form for different types of input objects. See Section 1.1.2.6 in
    Golyandina et al. (2018) for the explanation. Formal algorithms are described in this book in Algorithm 2.2 for
    1D-SSA, Algorithm 4.3 for MSSA, Algorithm 5.2 for 2D-SSA and Algorithm 5.6 for Shaped 2D-SSA.

    Fast implementation of reconstruction with the help of FFT is described in Korobeynikov (2010) for the 1D case
    and in Section 6.2 (Rank-one quasi-hankelization) of Golyandina et al. (2015) for the general case.


    Note
    ----

    By default (argument drop.attributes) the routine tries to preserve all the attributes of the input object. This
    way, for example, the reconstruction result of 'ts' object is the 'ts' object with the same time scale.

    References
    ----------

    Golyandina N., Korobeynikov A., Zhigljavsky A. (2018): Singular Spectrum Analysis with R. Use R!. Springer,
    Berlin, Heidelberg.

    Korobeynikov, A. (2010): Computation- and space-efficient implementation of SSA. Statistics and Its Interface,
    Vol. 3, No. 3, Pp. 257-268

    Golyandina, N., Korobeynikov, A., Shlemov, A. and Usevich, K. (2015): Multivariate and 2D Extensions of Singular
    Spectrum Analysis with the Rssa Package. Journal of Statistical Software, Vol. 67, Issue 2.
    doi:10.18637/jss.v067.i02

    Examples
    --------

    .. code-block:: python

        import pyrssa as prs
        import pandas as pd
        import numpy as np

        AustralianWine = prs.data("AustralianWine")
        fort = AustralianWine['Fortified'][:174]
        fort.index = pd.date_range(start='1980/01/01', freq='M', periods=len(fort))
        s_fort = prs.ssa(fort, L=84, kind="1d-ssa")
        r_fort = prs.reconstruct(s_fort, groups={"Trend": 1, "Seasonality": range(2, 13)})

        r_fort_2 = prs.reconstruct(s_fort, groups=[1, range(2, 13)])

        co2 = data("co2")
        s2 = prs.ssa(co2.value, column_projector="centering", row_projector="centering")
        rec = prs.reconstruct(s2, groups={"Linear_trend": range(1, s2.nspecial() + 1)})

    """
    return Reconstruction(x=x, groups=groups, drop_attributes=drop_attributes, cache=cache)


def iossa(x: SSA, nested_groups, tol=1e-5, kappa=2, maxiter=100, norm=None, trace=False, kappa_balance=0.5, **kwargs):
    return IOSSA(x=x, nested_groups=nested_groups, tol=tol, kappa=kappa, maxiter=maxiter, norm=norm, trace=trace,
                 kappa_balance=kappa_balance, call=_get_call(inspect.currentframe().f_back), **kwargs)


def fossa(x: SSA, nested_groups, filter=(-1, 1), gamma=np.inf, normalize=True, call=None, **kwargs):
    return FOSSA(x=x, nested_groups=nested_groups, filter=filter,
                 gamma=gamma, normalize=normalize, **kwargs, call=call)


# Weighted correlations
def wcor(ds, groups=range(1, 51)):
    return WCorMatrix(ds, groups)


# Forecasting functions

def rforecast(ds, groups, length=1, base="reconstructed", only_new=True, reverse=False,
              drop=False, drop_attributes=False, cache=True, **kwargs):
    return RForecast(ds, groups, length=length, base=base, only_new=only_new, reverse=reverse,
                     drop=drop, drop_attributes=drop_attributes, cache=cache, **kwargs)


def vforecast(ds, groups, length=1, only_new=True, drop=False, drop_attributes=False, **kwargs):
    return VForecast(ds, groups, length=length, only_new=only_new, drop=drop, drop_attributes=drop_attributes, **kwargs)


def bforecast(x, groups, length=1, R=100, level=0.95, kind="recurrent", interval="confidence",
              only_new=True, only_intervals=False, drop=True, drop_attributes=False, cache=True, **kwargs):
    return BForecast(x, groups, length=length, r=R, level=level, kind=kind, interval=interval, only_new=only_new,
                     only_intervals=only_intervals, drop=drop, drop_attributes=drop_attributes, cache=cache, **kwargs)


def hmatr(F, B=None, T=None, L=None, neig=10):
    return HMatrix(F, B=B, T=T, L=L, neig=neig)


def grouping_auto_pgram(x: SSABase, groups=None, base: Literal["series", "eigen", "factor"] = "series",
                        freq_bins=2, threshold=0, method: Literal["constant", "linear"] = "constant",
                        drop=True, **kwargs):
    return GroupPgram(x=x, groups=groups, base=base, freq_bins=freq_bins, threshold=threshold,
                      method=method, drop=drop, **kwargs)


def grouping_auto_wcor(x: SSABase, groups=None, nclust=None, **kwargs):
    return GroupWCor(x=x, groups=groups, nclust=nclust, **kwargs)


@overload
def grouping_auto(x: SSABase, grouping_method: Literal["pgram"] = "pgram", groups=None, nclust=None,
                  base: Literal["series", "eigen", "factor"] = "series", freq_bins=2, threshold=0,
                  method: Literal["constant", "linear"] = "constant", drop=True, **kwargs) -> GroupPgram:
    ...


@overload
def grouping_auto(x: SSABase, grouping_method: Literal["wcor"] = "wcor", groups=None, nclust=None,
                  base: Literal["series", "eigen", "factor"] = "series",
                  freq_bins=2, threshold=0,
                  method: Literal["ward.D", "ward.D2", "single", "complete",
                                  "average", "mcquitty", "median", "centroid"] = "complete",
                  drop=True, **kwargs) -> GroupWCor:
    ...


def grouping_auto(x: SSABase, grouping_method: str = "pgram", groups=None, nclust=None,
                  base: Literal["series", "eigen", "factor"] = "series", freq_bins=2, threshold=0,
                  method: Literal["constant", "linear"] = "constant", drop=True, **kwargs):
    if grouping_method == "pgram":
        return grouping_auto_pgram(x, groups=groups, base=base, freq_bins=freq_bins, threshold=threshold,
                                   method=method, drop=drop, **kwargs)
    elif grouping_method == "wcor":
        return grouping_auto_wcor(x, groups=groups, nclust=nclust, method=method, **kwargs)
    else:
        raise ValueError(f"Grouping method {grouping_method} is not in available methods: 'pgram', 'wcor'")


def spectrum(x, spans=None, kernel=None, taper=0.1,
             pad=0, fast=True, demean=False, detrend=True, **kwargs):
    return Periodogram(x, spans=spans, kernel=kernel, taper=taper,
                       pad=pad, fast=fast, demean=demean, detrend=detrend, **kwargs)
